package storage_test

import (
	"math/rand"
	"os"
	"path"
	"strconv"
	"testing"

	"codeberg.org/gruf/go-store/storage"
)

func TestDiskStorage(t *testing.T) {
	// Set test path, defer deleting it
	testPath := "diskstorage.test"
	t.Cleanup(func() {
		os.RemoveAll(testPath)
	})

	// Open new diskstorage instance
	st, err := storage.OpenFile(testPath, nil)
	if err != nil {
		t.Fatalf("Failed opening storage: %v", err)
	}

	// Attempt multi open of same instance
	_, err = storage.OpenFile(testPath, nil)
	if err == nil {
		t.Fatal("Successfully opened a locked storage instance")
	}

	// Run the storage tests
	// (this closes it)
	testStorage(t, st)

	// Test reopen storage path
	st, err = storage.OpenFile(testPath, nil)
	if err != nil {
		t.Fatalf("Failed opening storage: %v", err)
	}
	st.Close()
}

func TestDiskStorageDirTraversal(t *testing.T) {
	for _, path := range []string{
		"diskstorage.test",
		"./diskstorage.test",
		"/tmp/diskstorage.test",
		"/tmp/../tmp/diskstorage.test",
	} {
		testDiskStorageDirTraversal(t, path)
	}
}

func testDiskStorageDirTraversal(t *testing.T, path string) {
	t.Cleanup(func() {
		os.RemoveAll(path)
	})

	// Open new diskstorage instance
	st, err := storage.OpenFile(path, nil)
	if err != nil {
		t.Fatalf("Failed opening storage: %v", err)
	}
	defer st.Close()

	if _, err := st.Stat("../file.txt"); err != storage.ErrInvalidKey {
		t.Fatal("directory traversal did not result in error")
	}
	if _, err := st.Stat("./../file.txt"); err != storage.ErrInvalidKey {
		t.Fatal("directory traversal did not result in error")
	}
	if _, err := st.Stat("../../../file.txt"); err != storage.ErrInvalidKey {
		t.Fatal("directory traversal did not result in error")
	}
	if _, err := st.Stat("./../dir/../file.txt"); err != storage.ErrInvalidKey {
		t.Fatal("directory traversal did not result in error")
	}
}

func TestDiskStorageCustomLockfile(t *testing.T) {
	// Set test path, defer deleting it
	testPath := "diskstorage.test"
	lockfile := path.Join(os.TempDir(), tmpFileName())
	cfg := storage.DiskConfig{
		LockFile: lockfile,
	}
	t.Cleanup(func() {
		os.RemoveAll(testPath)
		os.Remove(lockfile)
	})

	// Open new diskstorage instance
	st, err := storage.OpenFile(testPath, &cfg)
	if err != nil {
		t.Fatalf("Failed opening storage: %v", err)
	}

	// Attempt multi open of same instance
	_, err = storage.OpenFile(testPath, &cfg)
	if err == nil {
		t.Fatal("Successfully opened a locked storage instance")
	}

	// Run the storage tests
	testStorage(t, st)

	// Test reopen storage path
	st, err = storage.OpenFile(testPath, &cfg)
	if err != nil {
		t.Fatalf("Failed opening storage: %v", err)
	}
	st.Close()
}

// tmpFileName generates a random temp file name.
func tmpFileName() string {
	return "tmp." + strconv.FormatInt(rand.Int63(), 10)
}
